package dataStructures;

import java.util.*;

/**
 * <p>Title: String Algorithms</p>
 * <p>Description: Includes a set of string algorithms</p>
 * @version 1.0
 */

class Pair {
  private int fst,snd;
  public Pair(int i,int j) { fst = i; snd = j; }
  public String toString() { return "(" + fst + "," + snd + ")"; }
}

public class StrAlgs {

  private StrAlgs() { } // no objects from this class

  /**
   * Search a pattern thru a text using brute-force, O(|P|*|T|)
   * @param P the pattern to find
   * @param T the text to be searched
   * @return a list of pairs defining the substrings of T equal to P
   */
  public static DList naiveSearch(String P, String T) {
     DList result = new DList();

     for(int i=T.length()-P.length();i>=0;i--) {
        boolean found = true;
        for (int j=0;found && j<P.length();j++)
          found = T.charAt(i+j) == P.charAt(j);
        if (found)
          result.addBegin(new Pair(i+1,i+P.length()));
     }

     return result;
  }

  /**
   * Search a pattern thru Z-boxes, O(|P|+|T|)
   *  Zi(S) is the length of the larger substrig S[i..] which is a prefix of S
   *
   * @requires (* P and T must not include any char '$' *)
   * @param P the pattern to find
   * @param T the text to be searched
   * @return a list of pairs defining the substrings of T equal to P
   */
  public static DList zBoxSearch(String P, String T) {
    String S = " "+P+"$"+T; // the initial space is to start at index 1
    DList result = new DList();

    int R=0,L=0; // the right and left limits of the rightmost Z-box
    int[] Z = new int[S.length()];

    // compute Z[2]
    Z[2] = getPrefixSize(S,1,2);
    if (Z[2]>0) {
      R = Z[2]+1;
      L = 2;
    }

    // compute Zk from the previous values
    for(int k=3;k<S.length();k++) {
      if (k>R) {  // k is outside the right-most Z-box: must perform match
        Z[k] = getPrefixSize(S,1,k);
        if (Z[k]>0) {
          R = k+Z[k]-1;
          L = k;
        }
      } else { // k<=R, ie, k is inside a Z-box
        int k_ = k-L+1;
        int betaSize = R-k+1;

        if (Z[k_] < betaSize) {
          Z[k] = Z[k_];
        } else {
          int q = (R+1) + getPrefixSize(S,R+1,betaSize+1);
          Z[k] = q-k;
          R = q-1;
          L = k;
        }
      }
    } //for(k)

    // All Zs are now computed!

    // Browse Z: each value = |P| is an initial index of an exact match
    for(int i=Z.length-1;i>P.length()+1;i--)
      if (Z[i]==P.length())
        result.addBegin(new Pair(i-(P.length()+1),i-2));

    return result;
  }

  /**
   * @param S The string to search
   * @param i The first index to start matching
   * @param j The second index to start matching
   * @return the length of the larger substrig S[j..] prefix of S[i..]
   */
  private static int getPrefixSize(String S, int i, int j) {
    int pos=0;
    while(j+pos<S.length() && i+pos<S.length() &&
          S.charAt(i+pos)==S.charAt(j+pos))
      pos++;
    return pos;
  }

  //////////////////////////////////////////////////////////////

  public static void main(String[] args) {
    String P = "aba", T = "abacdeabfababa";

    System.out.println(naiveSearch(P,T));
    System.out.println(zBoxSearch(P,T));
  }
}